from np_config import rigs


def test_rig_mvr_config():
    """Test that the mvr_config property of the Rig class raises no errors."""
    rigs.Rig(2).mvr_config.as_posix() 
    assert True


def test_rig_sync_config():
    """Test that the sync_config property of the Rig class raises no errors."""
    rigs.Rig(2).sync_config.as_posix() 
    assert True


def test_rig_camstim_config():
    """Test that the camstim_config property of the Rig class raises no errors."""
    rigs.Rig(2).camstim_config.as_posix() 
    assert True


from __future__ import annotations

import contextlib
import datetime
import doctest
import logging
import os
import pathlib
import re
import socket
import sys
from typing import Any, Hashable, Mapping, MutableMapping, TypeVar

from typing_extensions import Literal
from singledispatch import singledispatch

logger = logging.getLogger(__name__)

HOSTNAME = socket.gethostname().upper()

MutableMappingType = TypeVar("MutableMappingType", bound=MutableMapping)


def merge(
    base: MutableMappingType, update: Mapping
) -> MutableMappingType:
    """
    Utility function to do a deep merge on dictionaries. `base` will be modified, so deep
    copy first if the original needs to be preserved. This is a recursive function, so
    it merges nested dicts. If a value isn't a dict in both `base` and `update` the
    values won't be merged: the value from `update` will override that of `base`.
    
    From mpeconfig.
    
    >>> merge({0: {'a': False}, 1: False}, {0: {'a': True, 'b': True}, 1: True})
    {0: {'a': True, 'b': True}, 1: True}
    """
    for key, value in update.items():
        if isinstance(value, Mapping):
            if key not in base:
                base[key] = type(value)()  # For subclasses of dict
            merge(base[key], update[key])
        else:
            base[key] = value
    return base


def local_to_unc(host: str, path: str | pathlib.Path) -> pathlib.Path:
    """Generate a UNC path (for Windows) from host + local path.
    
    >>> local_to_unc("W10DT713843", pathlib.Path("C:\\ProgramData\\AIBS_MPE\\MVR\\data")).as_posix()
    '//W10DT713843/C/ProgramData/AIBS_MPE/MVR/data'
    
    >>> local_to_unc("W10DT713843", "C$/ProgramData/AIBS_MPE/MVR/data").as_posix()
    '//W10DT713843/C$/ProgramData/AIBS_MPE/MVR/data'
    """
    host = host.strip("\\/ ").upper()
    path = pathlib.Path(path).as_posix().replace(":", "")
    return normalize_path(f"//{host}/{path}")


def unc_to_local(path: str | pathlib.Path) -> pathlib.Path:
    """Try to convert UNC path with drive letter to local path (for Windows).
    
    >>> unc_to_local("//W10DT713843/C/ProgramData/AIBS_MPE/MVR/data").as_posix()
    'C:/ProgramData/AIBS_MPE/MVR/data'
    
    Path with no drive letter will raise an error:
    >>> unc_to_local("//w10dtsm18306/neuropixels_data")
    Traceback (most recent call last):
     ...
    ValueError: UNC path points to a host + share name (not host + drive letter): insufficient information to construct a local path from //w10dtsm18306/neuropixels_data
    """
    path: pathlib.Path = normalize_path(path)

    if path.as_posix()[0] != "/":
        raise ValueError(f"Path must be UNC (starting with a slash): {path.as_posix()}")

    # UNC path starts with host + drive letter or share
    # Path.parts merges UNC host and share, so we need to split manually
    parts = [_ for _ in path.as_posix().split("/") if _]
    drive = parts[1].strip("$")
    if len(drive) > 1:
        raise ValueError(
            f"UNC path points to a host + share name (not host + drive letter): insufficient information to construct a local path from {path.as_posix()}"
        )
    return normalize_path(f"{drive}:/" + "/".join(parts[2:]))


def normalize_path(path: str | pathlib.Path) -> pathlib.Path:
    """Normalize a path to a pathlib.Path object, starting with `//` if a network path.
    
    >>> normalize_path("\\W10DT713843\\c/ProgramData/AIBS_MPE/MVR/data").as_posix()
    '//W10DT713843/c/ProgramData/AIBS_MPE/MVR/data'
    """
    path = str(path)
    path = path.replace("\\", "/")  # makes the next parts simpler
    if path[0] == "/" and path[1] != path[0]:
        path = "/" + path
        path.replace(":", "")
    return pathlib.Path(path)


def local_or_unc_path(host: str, path: str | pathlib.Path) -> pathlib.Path:
    """Get a path that works from a location on the host computer. 
    
    - converts a UNC path to a local path if running on the host computer specified in the
    UNC path.
    - converts a local path to a UNC path if running on a different computer.
    """
    if HOSTNAME == host:
        with contextlib.suppress(ValueError):
            _ = unc_to_local(path)
            if _.exists():
                return _
        # otherwise path is not unc
        normalized = normalize_path(
            pathlib.Path(str(path).replace("$", ":"))
        )
        if not normalized.exists():
            logger.warning("Path not found in local filesystem: %s", path)
        return normalized

    if host not in str(path):
        return local_to_unc(host, path)
    return normalize_path(path)


def rig_idx(id: str | None) -> Literal[0, 1, 2, 3, 4] | None:
    """Convert rig ID (`'NP.1'`) to valid rig index (`1`).
    
    >>> rig_idx('NP.1')
    1
    >>> rig_idx('NP1')
    1

    Also accepts bad input, if index is valid:
    >>> rig_idx('1')
    1
    >>> rig_idx(1)
    1
    """
    if id is None:
        return None
    valid_idx = (0, 1, 2, 3, 4)
    with contextlib.suppress(Exception):
        result = int(str(id))  # if idx provided instead of id
        if result in valid_idx:
            return result
    result = re.findall(fr"NP\.?([\d]+)", id.upper())
    if not result:
        return None
    result = int(result[0])
    if result not in valid_idx:
        return None
    return result


PLATFORM_JSON_TIME_FMT = "%Y%m%d%H%M%S"

@singledispatch
def normalize_time(t: str | float | int | datetime.datetime) -> str: 
    """
    A standard time format matching platform JSON files.
    
    >>> normalize_time(datetime.datetime(2023, 2, 14, 13, 30, 00))
    '20230214133000'
    >>> normalize_time(1676410200.0)
    '20230214133000'
    >>> normalize_time(1676410200)
    '20230214133000'
    >>> normalize_time('1676410200.0')
    '20230214133000'
    >>> normalize_time('1676410200')
    '20230214133000'
    >>> normalize_time('2023-02-14T13:30:00')
    '20230214133000'
    """
    return _
    
@normalize_time.register
def _(t: datetime.datetime) -> str:
    return t.strftime(PLATFORM_JSON_TIME_FMT)

@normalize_time.register
def _(t: float) -> str:
    return datetime.datetime.fromtimestamp(t).strftime(PLATFORM_JSON_TIME_FMT)

@normalize_time.register
def _(t: int) -> str:
    # can't use int | float because of future.__annotations__
    return datetime.datetime.fromtimestamp(t).strftime(PLATFORM_JSON_TIME_FMT)

@normalize_time.register
def _(t: str) -> str:
    if len(t) == 14: # already in correct format
        return datetime.datetime(int(t[:4]), *(int(t[_:_+2]) for _ in range(4, 14, 2))).strftime(PLATFORM_JSON_TIME_FMT)
    with contextlib.suppress(ValueError):
        return datetime.datetime.fromtimestamp(float(t)).strftime(PLATFORM_JSON_TIME_FMT)
    with contextlib.suppress(ValueError):
        return datetime.datetime.fromisoformat(t).strftime(PLATFORM_JSON_TIME_FMT) 
    raise ValueError(f"Unable to parse time: {t}")

if __name__ == "__main__":
    doctest.testmod()



"""Access to rig computer hostnames and rig-wide ZooKeeper configs.

::
### When running on a rig-attached computer

AIBS MPE computer and rig IDs:
>>> COMP_ID, get_rig_id(), get_rig_idx()        # doctest: +SKIP
'NP.1-Acq', 'NP.1', 1


### For specific rigs

AIBS MPE rig ID:
>>> Rig(1).id
'NP.1'

Hostnames for each rig computer [Sync, Mon, Acq, Stim]:
>>> Rig(1).Acq
'W10DT713843'

Paths for specific apps running on a rig:
>>> Rig(1).paths['MVR'].as_posix()
'//W10DTSM118295/c$/ProgramData/AIBS_MPE/mvr/data'

Config dict for a particular rig, fetched from ZooKeeper /rigs/NP.<idx>:
>>> Rig(1).config['pretest_mouse']        
599657

When running on a rig, its NP-index is obtained from an env var, making the current rig's
properties available by default:
>>> Rig().Acq                       # doctest: +SKIP 
'W10DT713843'

If app is running on the local computer, its path is represented as a local path:
>>> Rig().paths['Sync']             # doctest: +SKIP
WindowsPath('C:/ProgramData/AIBS_MPE/sync/data')

...otherwise, as a network path:
>>> Rig().paths['Stim']             # doctest: +SKIP
WindowsPath('//W10DT713942/c$/ProgramData/AIBS_MPE/camstim/data')

"""
from __future__ import annotations

import doctest
import functools
import logging
import os
import pathlib
from typing import Any, Hashable, Optional

import requests
from backports.cached_property import cached_property
from typing_extensions import Literal

import np_config.config as config
import np_config.utils as utils

logger = logging.getLogger(__name__)

# all mpe computers --------------------------------------------------------------------

SERVER = "http://mpe-computers/v2.0"


@functools.lru_cache(maxsize=None)
def get_mpe_computer_ids(key: Literal['comp_ids', 'rig_ids', 'cluster_ids'] | None = None) -> dict[str, Any]:
    """Get a mapping of computer/rig/cluster IDs and their corresponding information."""
    all_ids = requests.get(SERVER).json()
    if key:
        return all_ids[key]
    return all_ids

get_comp_ids = functools.partial(get_mpe_computer_ids, "comp_ids")
get_rig_ids = functools.partial(get_mpe_computer_ids, "rig_ids")
get_cluster_ids = functools.partial(get_mpe_computer_ids, "cluster_ids")

# make mappings for easier lookup
def get_id_to_comp_ids() -> dict[str, list[str]]:
    "Keys are rig IDs (`NP.1`), values are lists of computer IDs (`['NP.1-Acq', ...]`)."
    return {k: v.get("comp_ids", []) for k, v in get_rig_ids().items()}

def get_comp_id_to_hostname() -> dict[str, str]:
    "Keys are computer IDs (`NP.1-Acq`), values are hostnames (`W10DT713843`)."
    return {k: v.get("hostname", "").upper() for k, v in get_comp_ids().items()}

def get_hostname_to_comp_id() -> dict[str, str]:
    "Keys are hostnames (`W10DT713843`), values are computer IDs (`NP.1-Acq`)."
    return {v.upper(): k for k, v in get_comp_id_to_hostname().items()}

def get_rig_id_to_hostnames() -> dict[str, list[str]]:
    "Keys are rig IDs (`NP.1`), values are lists of hostnames (`['W10DT713843', ...]`)."
    return {k: [get_comp_id_to_hostname()[comp_id] for comp_id in v] for k, v in get_id_to_comp_ids().items()}

# local computer properties ------------------------------------------------------------

@functools.lru_cache(maxsize=None)
def get_comp_id() -> str | None:
    "AIBS MPE comp ID for this computer, e.g. `NP.1-Sync`."
    return get_hostname_to_comp_id().get(utils.HOSTNAME) or os.environ.get("AIBS_COMP_ID") or None

@functools.lru_cache(maxsize=None)
def get_rig_id() -> str | None:
    "AIBS MPE NP-rig ID, e.g. `'NP.1'` if running on a computer connected to NP.1."
    return (
        os.environ.get("AIBS_RIG_ID", "").upper()
        or get_comp_ids().get(get_comp_id() or "", {}).get("rig_id")
        or (
            f"NP.{utils.rig_idx(get_comp_id())}"
            if get_comp_id() and utils.rig_idx(get_comp_id()) is not None
            else None
        )
        or ("BTVTest.1" if os.environ.get("USE_TEST_RIG", False) else None)
        or None
    )

def get_rig_idx() -> int | None:
    "AIBS MPE NP-rig index, e.g. `1` if running on a computer connected to NP.1."
    return utils.rig_idx(get_rig_id())

class Rig:
    """Access to rig computer hostnames and rig-wide ZooKeeper configs.
    ::
    
    AIBS MPE rig ID:
    >>> Rig(1).id
    'NP.1'

    Hostnames for each rig computer [Sync, Mon, Acq, Stim]:
    >>> Rig(1).Acq
    'W10DT713843'

    Config dict for a particular rig, fetched from ZooKeeper /rigs/NP.<idx>:
    >>> Rig(1).config['pretest_mouse']
    599657

    When running on a rig, its NP-index is obtained from an env var, making the current rig's
    properties available by default (equivalent to `Rig(get_rig_idx())`):
    >>> Rig().Acq                       # doctest: +SKIP 
    'W10DT713843'

    >>> Rig().config['Acq']             # doctest: +SKIP
    'W10DT713843'

    """

    id: str
    "AIBS MPE rig ID, e.g. `NP.1`"
    idx: int
    "AIBS MPE NP-rig index, e.g. `1` for NP.1"

    _sync: str
    _stim: str
    _mon: str
    _acq: str

    def __init__(self, idx_or_id: Optional[int | str] = None):
        idx_or_name = get_rig_idx() if idx_or_id is None else idx_or_id
        if idx_or_name is None:
            raise ValueError("Rig index not specified and not running on a rig.")
        idx = utils.rig_idx(str(idx_or_name))
        if idx is None:
            raise ValueError(f"`NP.{idx}` is not a recognized NP-rig")
        self.idx: int = idx
        self.id: str = f"NP.{idx}"
        for comp in ("sync", "stim", "mon", "acq"):
            setattr(self, f"_{comp}", get_comp_id_to_hostname()[f"{self.id}-{comp.title()}"])

    def __str__(self) -> str:
        return self.id

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id!r})"

    @property
    def sync(self) -> str:
        "Hostname for the Sync computer."
        return self._sync

    SYNC = Sync = sync

    @property
    def mon(self) -> str:
        "Hostname for the Mon computer."
        return self._mon

    MON = Mon = vidmon = VidMon = VIDMON = mon

    @property
    def acq(self) -> str:
        "Hostname for the Acq computer."
        return self._acq

    ACQ = Acq = acq

    @property
    def stim(self) -> str:
        "Hostname for the Stim computer."
        return self._stim

    STIM = Stim = stim

    @cached_property
    def config(self) -> dict[Hashable, Any]:
        "Rig-specific config dict, fetched from ZooKeeper."
        return utils.merge(
            config.from_zk("/np_defaults/configuration"),
            config.from_zk(f"/rigs/{self.id}"),
        )

    @property
    def paths(self) -> dict[str, pathlib.Path]:
        """Network paths to data folders for various devices/services, using 
        values from ZooKeeper /np_defaults/configuration and
        /rigs/NP.<idx>/paths.
        
        >>> Rig(1).paths['Sync'].as_posix()
        '//W10DT26AD0025/c$/ProgramData/AIBS_MPE/sync/data'
        """
        paths = dict()

        for service, service_config in self.config["services"].items():
            if "data" not in service_config:
                continue
            data_path = service_config["data"]
            host = getattr(self, service_config.get("comp"), None) or service_config["host"]
            if host in get_rig_id_to_hostnames()[self.id]:
                paths[str(service)] = utils.local_or_unc_path(
                    host=host, path=service_config["data"]
                )
            else:
                paths[str(service)] = utils.normalize_path(f"//{host}/{data_path}")
        for name, path in self.config.get("paths", {}).items():
            paths[str(name)] = utils.normalize_path(path)

        return paths

    @property
    def mvr_config(self) -> pathlib.Path:
        "Path to MVR config file for this rig."
        return utils.normalize_path(
            f"//{self.mon}/c$/ProgramData/AIBS_MPE/mvr/config/mvr.ini"
        )
    
    @property
    def sync_config(self) -> pathlib.Path:
        "Path to sync config file for this rig."
        return utils.normalize_path(
            f'//{self.sync}/c$/ProgramData/AIBS_MPE/sync/config/sync.yml'
        )

    @property
    def camstim_config(self) -> pathlib.Path:
        "Path to camstim config file for this rig."
        return utils.normalize_path(
            f'//{self.stim}/c$/ProgramData/AIBS_MPE/camstim/config/camstim.yml'
        )

def get_rig_config(idx_or_id: Optional[int | str] = None) -> dict[Hashable, Any] | None:
    """Get the rig-specific config dict for the specified rig index or ID, fetched
    from ZooKeeper. If no index or ID is provided, returns the config dict for the
    current rig, if running on an NP rig.
    
    >>> get_rig_config(1)['pretest_mouse']
    599657
    """
    if idx_or_id is None:
        return Rig().config if get_rig_idx() else None
    return Rig(idx_or_id).config


if __name__ == "__main__":
    doctest.testmod()



import contextlib
import dataclasses
import os
import pathlib
import random
from typing import Dict, Mapping

import pytest

import np_config

key = "test"
known_zk_keys = ("/mpe_defaults/logging", "/mpe_defaults/configuration")


@pytest.fixture(scope="function")
def test_dict() -> Dict[str, int]:
    return {key: random.randint(0, 100)}


@pytest.fixture(scope="module")
def test_zk():
    with np_config.ConfigZK() as zk:
        yield zk
    with contextlib.suppress(KeyError):
        zk[key] = {}


def test_zk_online():
    assert np_config.host_responsive(np_config.MINDSCOPE_SERVER)


def test_to_file_from_file(tmp_path, test_dict):
    for suffix in (".json", ".yaml"):
        file = tmp_path / f"test{suffix}"
        np_config.to_file(test_dict, file)
        assert np_config.from_file(file) == test_dict


def test_to_zk_from_zk(test_dict):
    np_config.to_zk(test_dict, path=key)
    assert np_config.from_zk(key) == test_dict
    np_config.to_zk({}, path=key)


def standard_dict_methods(dict, value):
    dict[key] = value
    assert dict[key] == value
    del dict[key]
    assert pytest.raises(KeyError, dict.__getitem__, key)
    assert dict.get(key) is None
    assert dict.get(key, 1) is 1


def test_zk_cls(test_zk, test_dict):
    standard_dict_methods(test_zk, test_dict)


def standard_file_dict_methods(dict, value):
    dict[key] = value
    assert np_config.from_file(dict.file)[key] == value
    del dict[key]
    assert np_config.from_file(dict.file) == {}
    # test non-string key - not valid .json
    if dict.file.suffix != ".json":
        dict[0] = value
        assert np_config.from_file(dict.file)[0] == value


def test_file_cls(tmp_path, test_dict):
    file = tmp_path / "test.json"
    test_file = np_config.ConfigFile(file)
    standard_dict_methods(test_file, test_dict)
    standard_file_dict_methods(test_file, test_dict)
    file = tmp_path / "test.yaml"
    test_file = np_config.ConfigFile(file)
    standard_file_dict_methods(test_file, test_dict)


def test_recorded_cls(tmp_path, test_dict):
    file = tmp_path / "test.yaml"
    test_recorded = np_config.RecordedZK(record_file=file)
    test_recorded[key] = test_dict
    assert test_recorded[key] == test_dict
    assert np_config.from_file(file)[key] == test_dict
    # items can't be deleted from record on file
    del test_recorded[key]
    assert np_config.from_file(file)[key] == test_dict
    # but should be deleted from the monitored dict
    assert pytest.raises(KeyError, test_recorded.__getitem__, key)

    test_recorded = np_config.RecordedZK(record_file=file)
    # previously-accessed items should be restored from file, if file is re-used
    assert np_config.from_file(file)[key] == test_dict
    # accessed item should be recorded on file
    _ = test_recorded[known_zk_keys[0]]
    assert np_config.from_file(file)[known_zk_keys[0]] == _
    # items not accessed should not be recorded on file
    assert known_zk_keys[1] not in np_config.from_file(file).keys()
    # record-keeping can be disabled
    with test_recorded.no_record():
        assert test_recorded[known_zk_keys[1]]
    assert known_zk_keys[1] not in np_config.from_file(file).keys()


def test_session_zk_record(tmp_path, test_dict):
    file = tmp_path / "test.yaml"
    # mod the path the record is written to
    np_config.SESSION_ZK_RECORD.record.file = file
    np_config.to_zk(test_dict, path=key)
    _ = np_config.SESSION_ZK_RECORD[key]
    assert (
        np_config.from_file(np_config.SESSION_ZK_RECORD.record.file)[key] == test_dict
    )


def test_backup_zk(tmp_path):
    backup = tmp_path / "test_backup.json"
    np_config.backup_zk(backup)
    assert backup.exists()
    server = np_config.ConfigZK()
    with server:
        assert all(server[paths] for paths in np_config.from_file(backup).keys())


def backed_up_zk(tmp_path, test_dict):
    backup = tmp_path / "test_backedup.json"
    np_config.to_file(test_dict, backup)
    test_zk = np_config.BackedUpZK(backup_file=backup)

    def using_local_backup(obj):
        return not hasattr(obj, "hosts")

    # known keys should be accessible either way
    assert test_zk[known_zk_keys[0]]

    # force local backup to be used:
    test_zk = np_config.BackedUpZK(hosts="wrong:0000", backup_file=backup)
    assert using_local_backup(test_zk)
    assert test_zk[known_zk_keys[0]]


# test_zk_backup()



from __future__ import annotations

import atexit
import collections
import contextlib
import datetime
import itertools
import json
import logging
import pathlib
import platform
import subprocess
import sys
import threading
from typing import Any, Generator, Mapping

import platformdirs
import yaml
from kazoo.client import KazooClient

# TODO use local backup for ZK if server unavailable

logging.getLogger('kazoo.client').setLevel('WARNING')
logger = logging.getLogger(__name__)

# preserve order of keys in dict
yaml.add_representer(
    dict,
    lambda self, data: yaml.representer.SafeRepresenter.represent_dict(
        self, data.items()
    ),
)

ZK_HOST_PORT: str = "eng-mindscope:2181"
MINDSCOPE_SERVER: str = "eng-mindscope.corp.alleninstitute.org"

ROOT_DIR: pathlib.Path = pathlib.Path(__file__).absolute().parent.parent
LOCAL_DATA_DIR = pathlib.Path(platformdirs.user_data_dir("np_config", "np"))
LOCAL_ZK_BACKUP_FILE = LOCAL_DATA_DIR / "zk_backup.yaml"
"File for keeping a full backup of Zookeeper configs."

session_start_time = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")
zk_record = (
    LOCAL_DATA_DIR / "config_logs" / f"zk_config-{pathlib.Path().cwd().name}.yaml"
)
SESSION_ZK_RECORD_FILE = zk_record.with_suffix(
    f".{session_start_time}{zk_record.suffix}"
)  # . is used later, don't change
"File for keeping a record of configs accessed from ZK during the current session."

SESSION_ZK_RECORD: "RecordedZK"  # created after class definition


def is_connected(host: str = MINDSCOPE_SERVER) -> bool:
    "Use OS's `ping` cmd to check if `host` is connected."
    command = ["ping", "-n" if "win" in sys.platform else "-c", "1", host]
    try:
        return subprocess.call(command, stdout=subprocess.PIPE, timeout=0.1) == 0
    except subprocess.TimeoutExpired:
        return False


def recorded_zk_config(**kwargs) -> "RecordedZK":
    return RecordedZK(record_file=SESSION_ZK_RECORD_FILE, **kwargs)


def cleanup_zk_records() -> None:
    "Remove current session zk record if it matches previous records, so we maintain config diffs only."
    path = SESSION_ZK_RECORD_FILE.parent

    def pairwise(iterable):
        "pairwise('ABCDEFG') --> AB BC CD DE EF FG"
        # itertools version not available in <3.10
        a, b = itertools.tee(iterable)
        next(b, None)
        return zip(a, b)

    for _, project_records in itertools.groupby(
        path.glob("*"), key=lambda f: f.stem.split(".")[0]
    ):
        for pair in pairwise(
            sorted(project_records, key=lambda f: f.stat().st_ctime, reverse=True)
        ):
            if pair[0].read_bytes() == pair[1].read_bytes():
                logger.debug(
                    f"Removing un-changed zk record: {pair[0].stem.split('.')[0]}"
                )
                with contextlib.suppress(Exception):
                    pair[0].unlink()
            break  # only compare against the next most recent record


atexit.register(cleanup_zk_records)


def from_zk(path: str, **kwargs) -> dict[Any, Any]:
    "Access eng-mindscope Zookeeper, return config dict."
    with ConfigZK(**kwargs) as zk:
        return zk[path]


def from_file(file: pathlib.Path) -> dict[Any, Any]:
    "Read file (yaml or json), return dict."
    file = pathlib.Path(file)
    with file.open("r") as f:
        if file.suffix in (".yaml", ".yml"):
            return yaml.load(f, Loader=yaml.loader.Loader) or dict()
        elif file.suffix == ".json":
            return json.load(f) or dict()
    raise ValueError(f"Config at {file} should be a .yaml or .json file.")


def normalize_zk_path(path: str) -> str:
    """
    >>> normalize_zk_path("project/config")
    '/project/config'
    >>> normalize_zk_path("\\\\project\\config")
    '/project/config'
    """
    path = str(path).replace("\\", "/")
    while path[0] != "/" or path[1] == "/":
        path = "/" + path.lstrip("/")
    return path


def fetch(arg: str | Mapping | pathlib.Path, **kwargs) -> dict[Any, Any]:
    "Differentiate a file path from a ZK path and return corresponding dict."

    if isinstance(arg, Mapping):
        config = arg

    elif isinstance(arg, (str, pathlib.Path)):
        # first rule-out that the output isn't a filepath
        path = pathlib.Path(str(arg)).resolve()
        if path.is_file() or path.suffix:
            config = from_file(path)

        elif isinstance(arg, str):
            # likely a ZK path
            path_str = normalize_zk_path(arg)
            config = from_zk(path_str, **kwargs)
    else:
        raise ValueError(
            "Logging config input should be a path to a .yaml or .json file, a ZooKeeper path, or a python logging config dict."
        )

    return dict(**config)


def to_file(config: dict, file: pathlib.Path):
    "Dump dict to file (yaml or json, based on file extension supplied)."
    file = pathlib.Path(file)
    file.parent.mkdir(parents=True, exist_ok=True)
    with file.open("w") as f:
        if file.suffix == ".yaml":
            return yaml.dump(config, f)
        elif file.suffix == ".json":
            return json.dump(config, f, indent=4, default=str)
    raise ValueError(f"Logging config {file} should be a .yaml or .json file.")


def to_zk(config: dict, path: str):
    "Dump config to Zookeeper at path, deleting path if config is empty."
    path = normalize_zk_path(path)
    with ConfigZK() as zk:
        if not config:
            del zk[path]
        else:
            zk[path] = config


def host_responsive(host: str) -> bool:
    """
    Remember that a host may not respond to a ping (ICMP) request even if the host name
    is valid. https://stackoverflow.com/a/32684938
    """
    param = "-n" if platform.system().lower() == "windows" else "-c"
    command = ["ping", param, "1", host]
    return subprocess.call(command, stdout=subprocess.PIPE) == 0


def backup_zk(file: pathlib.Path = LOCAL_ZK_BACKUP_FILE):
    "Recursively backup all zookeeper records to local file."
    zk = ConfigZK()
    backup = dict()

    def get(zk: ConfigZK, parent="/"):
        children = zk.get_children(parent)
        if not children:
            return
        for child in children:
            key = parent + child if parent == "/" else "/".join([parent, child])
            try:
                value = zk[key]
            except KeyError:
                continue
            if value:
                backup[key] = value
            else:
                get(zk, key)

    with zk:
        get(zk)
    to_file(backup, file)


class ConfigFile(collections.UserDict):
    """
    A dictionary wrapper around a continuously-synced serialized local copy of a config.

    Used for keeping a full backup of all configs on zookeeper, or for keeping a record
    of the config fetched during a session.
    """

    lock: threading.Lock = threading.Lock()
    read_only: bool = False

    def __init__(self, file: pathlib.Path = SESSION_ZK_RECORD_FILE, dict=None):
        self.file = pathlib.Path(file)
        if not dict and self.file.exists():
            dict = from_file(self.file)
        super().__init__(dict)

    def write(self):
        if self.read_only:
            logger.debug("Not writing to read-only config file %s", self.file)
            return
        if not self.file.exists():
            self.file.parent.mkdir(parents=True, exist_ok=True)
            self.file.touch(exist_ok=True)
        with self.lock, contextlib.suppress(Exception):
            to_file(self.data, self.file)
            logger.debug(f"Updated local config file {self.file}")
            return
        logger.debug(
            f"Could not update local config file {self.file}", exc_info=True,
        )

    def __getitem__(self, key: Any):
        logger.debug(f"Fetching {key!r} from local config backup")
        try:
            value = super().__getitem__(key)
        except Exception as exc:
            raise KeyError(
                f"{key!r} not found in local config file {self.file}"
            ) from exc
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.write()
        logger.debug(f"{key!r} updated in local config file {self.file}")

    def __delitem__(self, key: Any):
        try:
            super().__delitem__(key)
        except Exception as exc:
            raise KeyError(
                f"{key!r} not found in local config file {self.file}"
            ) from exc
        else:
            self.write()
            logger.debug(f"{key!r} deleted from local config file {self.file}")

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.write()


class ConfigZK(KazooClient):
    def __init__(self, hosts: str = ZK_HOST_PORT, *args, **kwargs):
        """
        A dictionary and context-manager wrapper around the zookeeper interface - modified from
        mpeconfig ConfigServer.
        """
        super().__init__(hosts, timeout=10)

    @contextlib.contextmanager
    def _start(self) -> Generator["ConfigZK", None, None]:
        "Start the client via contextmanager if not already started."
        # contextlib.nullcontext preferred in 3.7+
        if not self.connected:
            with self:
                yield self
            self.__exit__()
        else:
            yield self

    def get(self, key: Any, default: Any = None) -> Any:
        "Overload KazooClient.get())"
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key) -> dict:
        with self._start():
            try:
                node = super().get(key)
            except Exception as exc:
                raise KeyError(f"{key!r} not found in zookeeper.") from exc
            else:
                item = node[0] if node and len(node) else None
                value = yaml.load(item or "", Loader=yaml.loader.Loader) or dict()
                return value

    def __setitem__(self, key, value):
        with self._start():
            self.ensure_path(key)
            super().set(key, bytes(yaml.dump(value or dict()), "utf-8"))

    def __delitem__(self, key):
        with self._start():
            try:
                super().delete(key)
            except Exception as exc:
                raise KeyError(f"{key!r} not found in zookeeper.") from exc

    def __enter__(self):
        if self.connected:
            return self
        try:
            self.start(timeout=10)
        except Exception as exc:
            if not is_connected():
                logger.debug(
                    "Starting the Kazoo client failed: %s", self.hosts, exc_info=True
                )
                raise ConnectionError(
                    f"Zookeeper server is unreachable: {MINDSCOPE_SERVER}"
                ) from exc
            raise exc
        else:
            return self

    def __exit__(self, *args, **kwargs):
        self.stop()


class RecordedZK(ConfigZK):

    skip_record = False
    record: ConfigFile
    "Dict class that records accessed keys on file."

    def __init__(self, record_file=SESSION_ZK_RECORD_FILE, *args, **kwargs):
        """
        A ZK config dict wrapper that records all keys accessed during a session to file.
        """
        self.record = ConfigFile(record_file)
        with self.no_record():
            super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        value = super().__getitem__(key)
        if not self.skip_record:
            self.record[key] = value
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if not self.skip_record:
            self.record[key] = value

    @contextlib.contextmanager
    def no_record(self) -> Generator[None, None, None]:
        logger.debug("Temporarily pausing record keeping %s.", self.__class__)
        self.skip_record = True
        try:
            yield
        finally:
            self.skip_record = False
            logger.debug("Resuming record keeping %s.", self.__class__)


SESSION_ZK_RECORD: RecordedZK = recorded_zk_config()


class BackedUpZK(ConfigZK, ConfigFile):
    def __init__(
        self, backup_file: pathlib.Path = LOCAL_ZK_BACKUP_FILE, *args, **kwargs
    ):
        """
        Zookeeper wrapper that pulls from local backup if server is not responsive.
        """
        test = ConfigZK(*args, **kwargs)
        try:
            _ = test.get_children("/")
        except:
            logger.debug("Could not connect to Zookeeper, using local backup file.")
            self.__class__ = ConfigFile
            ConfigFile.__init__(self, file=backup_file, *args, **kwargs)
            self.read_only = True

        self.__class__ = ConfigZK
        ConfigZK.__init__(self, *args, **kwargs)


# class Backedup(ConfigZK):

#     config: ConfigZK()
#     backup: ConfigFile

#     def __init__(self, backup_file: pathlib.Path = LOCAL_ZK_BACKUP_FILE, *args, **kwargs):
#         """
#         Zookeeper wrapper that pulls from local backup if server is not responsive.
#         """
#         self.backup = ConfigFile(backup_file)
#         with self.backup():
#             self.config.__init__(*args, **kwargs)

#     @contextlib.contextmanager
#     def backup(self) -> Generator[None, None, None]:
#         if not host_responsive(MINDSCOPE_SERVER):
#             logger.debug("Could not connect to Zookeeper, using local backup file.")
#             self.config = __class__.backup
#         yield
#         self.config = __class__.config

#     def __enter__(self):
#         with self.backup():
#             return self.config.__enter__()

#     def __exit__(self, *args, **kwargs):
#         with self.backup():
#             return self.config.__exit__(*args, **kwargs)

#     def __getitem__(self, key: Any):
#         with self.backup():
#             return self.config.__getitem__(key)

#     def __setitem__(self, key, value):
#         with self.backup():
#             self.config.__setitem__(key, value)

# ZK_BACKUP = ConfigFile(LOCAL_ZK_BACKUP_FILE)

if __name__ == "__main__":
    import doctest

    doctest.testmod()
